"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org


Description:

Functions:

"""

import os
import sys
import copy
import time

import cv2
import pyacvd
import numpy as np
import pyvista as pv
import SimpleITK as sitk

import vtk
from vtk.util import numpy_support

from scipy.spatial.distance import cdist
from matplotlib import path


class ContourToDiscreteMesh(object):
    """
    Converts 2D contour points into a 3D discrete or smoothed surface mesh, optionally generating a volumetric mask
    along the way. It supports multiple anatomical planes (Axial, Coronal, Sagittal).
    """
    def __init__(self, contour_position=None, contour_pixel=None, spacing=None, origin=None, dimensions=None,
                 matrix=None, plane='Axial', mask=None):
        """
        contour_position: Original contour points in physical coordinates.
        contour_pixel: Contours converted to voxel coordinates.
        spacing, origin, dimensions: Image geometry information.
        matrix: Transformation matrix for voxel conversion.
        plane: Plane along which slices are defined (Axial, Coronal, Sagittal).
        """
        self.contour_position = contour_position
        self.contour_pixel = contour_pixel
        self.spacing = spacing
        self.origin = origin
        self.dimensions = dimensions
        self.plane = plane

        self.mask = mask

        if matrix is None:
            self.matrix = np.identity(3)
        else:
            self.matrix = matrix

        if self.contour_pixel is None and self.mask is None:
            self.convert_to_pixel_spacing()

        if self.mask is None:
            self.compute_mask()

    def convert_to_pixel_spacing(self):
        """
        Converts physical contour points to voxel coordinates using a 4x4 conversion matrix.
        """
        matrix = np.identity(3, dtype=np.float32)
        matrix[0, :] = self.matrix[0, :] / self.spacing[0]
        matrix[1, :] = self.matrix[1, :] / self.spacing[1]
        matrix[2, :] = self.matrix[2, :] / self.spacing[2]

        conversion_matrix = np.identity(4, dtype=np.float32)
        conversion_matrix[:3, :3] = matrix
        conversion_matrix[:3, 3] = np.asarray(self.origin).dot(-matrix.T)

        self.contour_pixel = []
        for ii, pos in enumerate(self.contour_position):
            p_concat = np.concatenate((pos, np.ones((pos.shape[0], 1))), axis=1)
            self.contour_pixel += [p_concat.dot(conversion_matrix.T)[:, :3]]

    def compute_mask(self):
        """
        Generates a 3D binary mask slice-by-slice.
        """
        hold_mask = np.zeros([self.dimensions[0], self.dimensions[1], self.dimensions[2]], dtype=np.uint8)
        if self.plane == 'Axial':
            slice_check = np.zeros(self.dimensions[0])
            for c in self.contour_pixel:
                contour_stacked = np.vstack((c[:, 0:2], c[0, 0:2]))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[1], self.dimensions[2]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 2]))
                if slice_check[slice_num] == 0:
                    hold_mask[slice_num, :, :] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[slice_num, :, :] = hold_mask[slice_num, :, :] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

        elif self.plane == 'Coronal':
            slice_check = np.zeros(self.dimensions[1])
            for c in self.contour_pixel:
                stack_1 = np.vstack((c[:, 0], c[:, 2])).T
                stack_2 = np.asarray([c[0, 0], c[0, 2]])
                contour_stacked = np.vstack((stack_1, stack_2))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[0], self.dimensions[2]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 1]))
                if slice_check[slice_num] == 0:
                    hold_mask[:, slice_num, :] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[:, slice_num, :] = hold_mask[:, slice_num, :] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

        else:
            slice_check = np.zeros(self.dimensions[2])
            for c in self.contour_pixel:
                contour_stacked = np.vstack((c[:, 1:], c[0, 1:]))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[0], self.dimensions[1]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 0]))
                if slice_check[slice_num] == 0:
                    hold_mask[:, :, slice_num] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[:, :, slice_num] = hold_mask[:, :, slice_num] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

    def compute_mesh(self, discrete=False, smoothing_iterations=20, smoothing_relaxation=.5, smoothing_distance=1):
        """
        Converts the mask into a 3D mesh.
        """
        label = numpy_support.numpy_to_vtk(num_array=self.mask.ravel(),
                                           deep=True,
                                           array_type=vtk.VTK_FLOAT)

        img_vtk = vtk.vtkImageData()
        img_vtk.SetDimensions(self.dimensions[2], self.dimensions[1], self.dimensions[0])
        img_vtk.SetSpacing(self.spacing)
        img_vtk.SetOrigin(self.origin)
        img_vtk.SetDirectionMatrix(self.matrix.flatten(order='F'))
        img_vtk.GetPointData().SetScalars(label)

        pad_filter = vtk.vtkImageConstantPad()
        pad_filter.SetInputData(img_vtk)
        pad_filter.SetConstant(0)

        old_extent = img_vtk.GetExtent()
        new_extent = [old_extent[0] - 1, old_extent[1] + 1,
                      old_extent[2] - 1, old_extent[3] + 1,
                      old_extent[4] - 1, old_extent[5] + 1]
        pad_filter.SetOutputWholeExtent(new_extent)
        pad_filter.Update()
        pad_image = pad_filter.GetOutput()

        if discrete:
            vtk_mesh = vtk.vtkDiscreteMarchingCubes()
            vtk_mesh.SetInputData(pad_image)
            vtk_mesh.GenerateValues(1, 1, 1)
            vtk_mesh.Update()

            return pv.PolyData(vtk_mesh.GetOutput())

        else:
            # Uses VTK surface nets 3d
            img = pv.ImageData(pad_image)
            mesh = img.contour_labels(smoothing=True,
                                      output_mesh_type='triangles',
                                      smoothing_iterations=smoothing_iterations,
                                      smoothing_relaxation=smoothing_relaxation,
                                      smoothing_distance=smoothing_distance)

            return mesh


class ContourToMask(object):
    """
    Converts 2D contour points (from medical images or 3D scans) into a 3D volumetric mask, supporting multiple planes
    (Axial, Coronal, Sagittal).
    """
    def __init__(self, contour_position=None, contour_pixel=None, spacing=None, origin=None, dimensions=None,
                 matrix=None, plane='Axial'):
        """
        contour_position: Original contour points in physical coordinates.
        contour_pixel: Contours converted to pixel indices (voxel coordinates).
        spacing, origin, dimensions: Image geometry information.
        matrix: Transformation matrix to map physical coordinates to pixel space.
        plane: Axis along which slices are interpreted (Axial/Coronal/Sagittal).
        """
        self.contour_position = contour_position
        self.contour_pixel = contour_pixel
        self.spacing = spacing
        self.origin = origin
        self.dimensions = dimensions
        self.matrix = matrix
        self.plane = plane

        self.mask = None

    def create_mask(self):
        """
        Converts contours to pixel space if necessary, then computes the 3D mask.
        """
        if self.contour_pixel is None:
            self.convert_to_pixel_spacing()

        self.compute_mask()

    def convert_to_pixel_spacing(self):
        """
        Converts physical contour coordinates to voxel indices using SimpleITK.
        """
        sitk_image = sitk.Image([int(dim) for dim in self.dimensions], sitk.sitkUInt8)
        matrix_flat = self.matrix[0:3, 0:3].flatten(order='F')
        sitk_image.SetDirection([float(mat) for mat in matrix_flat])
        sitk_image.SetOrigin(self.origin)
        sitk_image.SetSpacing(self.spacing)

        self.contour_pixel = [[]] * len(self.contour_position)
        for ii, contours in enumerate(self.contour_position):
            self.contour_pixel[ii] = [sitk_image.TransformPhysicalPointToContinuousIndex(contour) for contour in contours]

    def compute_mask(self):
        """
        Generates a 3D mask array from the pixel contours.
        """
        hold_mask = np.zeros([self.dimensions[0], self.dimensions[1], self.dimensions[2]], dtype=np.uint8)
        if self.plane == 'Axial':
            slice_check = np.zeros(self.dimensions[0])
            for c in self.contour_pixel:
                contour_stacked = np.vstack((c[:, 1:], c[0, 1:]))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[1], self.dimensions[2]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 0]))
                if slice_check[slice_num] == 0:
                    hold_mask[slice_num, :, :] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[slice_num, :, :] = hold_mask[slice_num, :, :] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

        elif self.plane == 'Coronal':
            slice_check = np.zeros(self.dimensions[1])
            for c in self.contour_pixel:
                stack_1 = np.vstack((c[:, 0], c[:, 2])).T
                stack_2 = np.asarray([c[0, 0], c[0, 2]])
                contour_stacked = np.vstack((stack_1, stack_2))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[0], self.dimensions[2]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 1]))
                if slice_check[slice_num] == 0:
                    hold_mask[:, slice_num, :] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[:, slice_num, :] = hold_mask[:, slice_num, :] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

        else:
            slice_check = np.zeros(self.dimensions[2])
            for c in self.contour_pixel:
                contour_stacked = np.vstack((c[:, 1:], c[0, 1:]))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[0], self.dimensions[1]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 0]))
                if slice_check[slice_num] == 0:
                    hold_mask[:, :, slice_num] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[:, :, slice_num] = hold_mask[:, :, slice_num] + image
            self.mask = (hold_mask > 0).astype(np.uint8)


class ModelToMask(object):
    """
    Converts 3D model(s) into a volumetric mask. The mask can either be:
        - an empty array (default)
        - a binary mask representing the models.

    The mask is padded with a spacing buffer of 5 units along each axis (x, y, z) to ensure the model fits comfortably.
    """
    def __init__(self, models, origin=None, spacing=None, dims=None, slice_locations=None, matrix=None,
                 empty_array=True, convert=True):
        """
        Parameters
        ----------
        models : list
            List of 3D models to convert into a mask.
        origin : list or None
            The origin of the mask in x, y, z coordinates.
        spacing : list or None
            Voxel spacing along x, y, z axes.
        dims : list or None
            Dimensions of the mask array.
        slice_locations : list or None
            Z-axis slice locations at which contours are computed.
        matrix : np.array or None
            Transformation matrix for the models. Defaults to identity if None.
        empty_array : bool
            If True, only an empty mask is created. If False, the mask is filled with model contours.
        convert : bool
            If True, automatically computes bounds, contours, and mask during initialization.
        """
        self.models = models
        self.empty_array = empty_array

        self.spacing = spacing
        self.origin = origin
        self.dims = dims
        self.slice_locations = slice_locations

        if matrix is None:
            self.matrix = np.identity(4)
        else:
            self.matrix = matrix

        self.bounds = None

        self.contours = []
        self.mask = None

        if convert:
            self.compute_bounds()
            self.compute_contours()
            self.compute_mask()

    def compute_bounds(self):
        """
        Computes the bounding box for all models combined. Adds a buffer of 5 units multiplied by the spacing on each
        axis. Sets default spacing based on model size if spacing is not provided.
        """
        model_bounds = [model.GetBounds() for model in self.models]
        model_min = np.min(model_bounds, axis=0)
        model_max = np.max(model_bounds, axis=0)

        model_min_max = [model_min[0], model_max[1], model_min[2], model_max[3], model_min[4], model_max[5]]

        if model_min_max[1] - model_min_max[0] < 512 and model_min_max[3] - model_min_max[2] < 512:
            if model_min_max[5] - model_min_max[4] < 450:
                self.spacing = [1, 1, 3]

            elif model_min_max[5] - model_min_max[4] < 750:
                self.spacing = [1, 1, 5]

        if self.spacing is not None:
            self.bounds = [int(model_min_max[0] - 5 * self.spacing[0]), int(model_min_max[1] + 5 * self.spacing[0]),
                           int(model_min_max[2] - 5 * self.spacing[1]), int(model_min_max[3] + 5 * self.spacing[1]),
                           int(model_min_max[4] - 5 * self.spacing[2]), int(model_min_max[5] + 5 * self.spacing[2])]
            self.origin = [self.bounds[0], self.bounds[2], self.bounds[4]]

            self.slice_locations = [i for i in range(self.bounds[4], self.bounds[5], self.spacing[2])]
            self.dims = [len(self.slice_locations), self.bounds[3] - self.bounds[2] + 1, self.bounds[1] - self.bounds[0] + 1]

    def compute_contours(self):
        """
        Computes 2D contours along the z-axis for each model at the pre-defined slice locations. These contours are
        later used to fill the mask.
        """
        for model in self.models:
            com = model.center
            org_bounds = model.GetBounds()

            model_contours = []
            for s in self.slice_locations:
                if org_bounds[4] < s < org_bounds[5]:
                    hold_contour = model.slice(normal='z', origin=[com[0], com[1], s])
                    model_contours.append((np.asarray(hold_contour.points)[:, 0:2] -
                                     (self.bounds[0], self.bounds[2])) / (self.spacing[0:2]))
                else:
                    model_contours.append([])

            self.contours.append(model_contours)

    def compute_mask(self):
        """
        Creates a 3D mask. Default is an empty array. If empty_array=False, fills mask using computed contours.
        """
        self.mask = np.zeros((self.dims[0], self.dims[1], self.dims[2]))
        if not self.empty_array:
            for ii, model in enumerate(self.models):

                model_contours = self.contours[ii]
                for jj, s in enumerate(self.slice_locations):
                    if len(model_contours[jj]) > 0:
                        frame = np.zeros((self.dims[1], self.dims[2]))
                        cv2.fillPoly(frame, np.array([model_contours[jj]], dtype=np.int32), 1)
                        self.mask[jj, :, :] = self.mask[jj, :, :] + frame

        self.mask = self.mask.astype(np.int8)

    def save_image(self, export_path):
        """
        Saves the mask as an image file using SimpleITK.

        Parameters
        ----------
        export_path : str
            Output file path to save the mask.

        """
        image = sitk.GetImageFromArray(self.mask)
        image.SetSpacing(self.spacing)
        image.SetOrigin([self.bounds[0], self.bounds[2], self.bounds[4]])
        sitk.WriteImage(image, export_path)



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org


Description:

Functions:

"""


class Data(object):
    """
    Central storage class for managing images, rigid transformations, deformable transformations, doses, and associated
    objects like ROIs (regions of interest) and POIs (points of interest).

    This class uses class-level attributes to store all data and provides methods to maintain consistency across
    multiple images.
    """

    image = {}
    rigid = {}
    deformable = {}
    dose = {}

    image_list = []
    deformable_list = []
    dose_list = []
    poi_list = []
    rigid_list = []
    roi_list = []

    @classmethod
    def clear(cls):
        """
        Resets all data structures and lists. Clears all images, transformations, doses, ROIs, and POIs.
        """
        cls.image = {}
        cls.rigid = {}
        cls.deformable = {}
        cls.dose = {}

        cls.image_list = []
        cls.poi_list = []
        cls.roi_list = []
        cls.rigid_list = []

    @classmethod
    def delete_image(cls, image_name):
        """
        Deletes a specific image and removes it from the image list.

        Parameters
        ----------
        image_name : str
            Name of the image to remove.
        """
        del cls.image[image_name]
        del cls.image_list[image_name]

    @classmethod
    def match_rois(cls):
        """
        Ensures that all images contain the same ROIs (regions of interest).
        - Collects all ROI names from all images.
        - Synchronizes color and visibility of ROIs across images.
        - Adds missing ROIs to images with default or matched properties.
        """
        image_rois = [list(cls.image[image_name].rois.keys()) for image_name in list(cls.image.keys())]
        roi_names = list({x for r in image_rois for x in r})
        Data.roi_list = roi_names

        color = [[128, 128, 128]] * len(roi_names)
        visible = [False] * len(roi_names)
        for ii, roi_name in enumerate(roi_names):
            for image_name in list(cls.image.keys()):
                rois_on_image = list(cls.image[image_name].rois.keys())
                if roi_name in rois_on_image:
                    if cls.image[image_name].rois[roi_name].color is not None:
                        color[ii] = cls.image[image_name].rois[roi_name].color
                        visible[ii] = cls.image[image_name].rois[roi_name].visible

        for ii, roi_name in enumerate(roi_names):
            for image_name in list(cls.image.keys()):
                rois_on_image = list(cls.image[image_name].rois.keys())
                if roi_name not in rois_on_image:
                    cls.image[image_name].add_roi(roi_name=roi_name, color=color[ii], visible=visible[ii])

    @classmethod
    def match_pois(cls):
        """
        Ensures that all images contain the same POIs (points of interest).
        - Collects all POI names from all images.
        - Synchronizes color and visibility of POIs across images.
        - Adds missing POIs to images with default or matched properties.
        """
        image_pois = [list(cls.image[image_name].pois.keys()) for image_name in list(cls.image.keys())]
        poi_names = list({x for r in image_pois for x in r})
        Data.poi_list = poi_names

        color = [[128, 128, 128]] * len(poi_names)
        visible = [False] * len(poi_names)
        for ii, poi_name in enumerate(poi_names):
            for image_name in list(cls.image.keys()):
                pois_on_image = list(cls.image[image_name].pois.keys())
                if poi_name in pois_on_image:
                    if cls.image[image_name].pois[poi_name].color is not None:
                        color[ii] = cls.image[image_name].pois[poi_name].color
                        visible[ii] = cls.image[image_name].pois[poi_name].visible

        for ii, poi_name in enumerate(poi_names):
            for image_name in list(cls.image.keys()):
                pois_on_image = list(cls.image[image_name].pois.keys())
                if poi_name not in pois_on_image:
                    cls.image[image_name].add_poi(poi_name=poi_name, color=color[ii], visible=visible[ii])



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:

Structure:

"""

import copy

import numpy as np
import SimpleITK as sitk

import vtk
from vtkmodules.util import numpy_support
from scipy.ndimage import map_coordinates

from ..data import Data
from ..utils.deformable.simpleitk import DeformableITK


class Display(object):
    def __init__(self, deformable):
        self.deformable = deformable

        self.origin = None
        self.spacing = None
        self.array = None
        self.image = None
        self.matrix = np.identity(3)

        self.slice_location = [0, 0, 0]
        self.scroll_max = None
        self.offset = {'Axial': [0, 0], 'Coronal': [0, 0], 'Sagittal': [0, 0]}
        self.misc = {}

        self.compute_scroll_max()

    def compute_array(self, slice_plane):
        array_slice = None
        if slice_plane == 'Axial':
            if 0 <= self.slice_location[0] < self.array.shape[0]:
                array_slice = self.array[self.slice_location[0], :, :].astype(np.double)

        elif slice_plane == 'Coronal':
            if 0 <= self.slice_location[1] < self.array.shape[1]:
                array_slice = self.array[:, self.slice_location[1], :].astype(np.double)

        else:
            if 0 <= self.slice_location[2] < self.array.shape[2]:
                array_slice = self.array[:, :, self.slice_location[2]].astype(np.double)

        return array_slice

    def compute_deformation(self):
        image = self.deformable.create_image()

        self.array = sitk.GetArrayFromImage(image)
        self.spacing = image.GetSpacing()
        self.origin = image.GetOrigin()

    def compute_grid(self, slice_plane='Axial', vector='x'):
        if slice_plane == 'Axial':
            dvf_plane = self.deformable.dvf[self.slice_location[0], :, :, :]
        elif slice_plane == 'Coronal':
            dvf_plane = self.deformable.dvf[:, self.slice_location[1], :, :]
        else:
            dvf_plane = self.deformable.dvf[:, :, self.slice_location[2], :]

        if vector == 'x':
            dvf_vector = dvf_plane[:, :, 0]
        elif vector == 'y':
            dvf_vector = dvf_plane[:, :, 1]
        else:
            dvf_vector = dvf_plane[:, :, 2]

        return dvf_vector.astype(np.float32)

    def compute_matrix_pixel_to_position(self):
        matrix = copy.deepcopy(self.matrix)

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * self.spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * self.spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * self.spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        return pixel_to_position_matrix

    def compute_matrix_position_to_pixel(self):
        matrix = copy.deepcopy(self.matrix)

        hold_matrix = np.identity(3, dtype=np.float32)
        hold_matrix[0, :] = matrix[0, :] / self.spacing[0]
        hold_matrix[1, :] = matrix[1, :] / self.spacing[1]
        hold_matrix[2, :] = matrix[2, :] / self.spacing[2]

        position_to_pixel_matrix = np.identity(4, dtype=np.float32)
        position_to_pixel_matrix[:3, :3] = hold_matrix
        position_to_pixel_matrix[:3, 3] = np.asarray(self.origin).dot(-hold_matrix.T)

        return position_to_pixel_matrix

    def compute_mesh_slice(self, roi_name=None, location=None, slice_plane=None, return_pixel=False):
        if self.deformable.rois[roi_name] is None:
            self.deformable.update_rois(roi_name=roi_name)

    def compute_offset(self):
        if self.deformable.reference_name is not None:
            pos = Data.image[self.deformable.reference_name].origin

            self.offset['Axial'][0] = (self.origin[0] - pos[0]) / self.spacing[0]
            self.offset['Axial'][1] = (self.origin[1] - pos[1]) / self.spacing[1]
            self.offset['Coronal'][0] = (self.origin[0] - pos[0]) / self.spacing[0]
            self.offset['Coronal'][1] = (self.origin[2] - pos[2]) / self.spacing[2]
            self.offset['Sagittal'][0] = (self.origin[1] - pos[1]) / self.spacing[1]
            self.offset['Sagittal'][1] = (self.origin[2] - pos[2]) / self.spacing[2]

    def compute_slice_location(self, position=None):
        if position is None:
            source_location = np.flip(Data.image[self.deformable.reference_name].display.slice_location)
            position = Data.image[self.deformable.reference_name].display.compute_index_positions(source_location)
        self.slice_location = np.flip(np.round((position - self.origin) / self.spacing).astype(np.int32))

    def compute_slice_origin(self, slice_plane):
        slice_origin = None
        if slice_plane == 'Axial' and 0 <= self.slice_location[0] <= self.scroll_max[0]:
            location = np.asarray([0, 0, self.slice_location[0]])
            slice_origin = self.origin + (location * self.spacing)
        elif slice_plane == 'Coronal' and 0 <= self.slice_location[1] <= self.scroll_max[1]:
            location = np.asarray([0, self.slice_location[1], 0])
            slice_origin = self.origin + (location * self.spacing)
        elif slice_plane == 'Sagittal' and 0 <= self.slice_location[2] <= self.scroll_max[2]:
            location = np.asarray([self.slice_location[2], 0, 0])
            slice_origin = self.origin + (location * self.spacing)

        return slice_origin

    def compute_scroll_max(self):
        if self.array is None:
            self.scroll_max = self.deformable.dimensions - 1
        else:
            self.scroll_max = [self.array.shape[0] - 1,
                               self.array.shape[1] - 1,
                               self.array.shape[2] - 1]

    def convert_position_to_pixel(self, position=None):
        position_to_pixel_matrix = self.compute_matrix_position_to_pixel()

        pixel = []
        for ii, pos in enumerate(position):
            p_concat = np.concatenate((pos, np.ones((pos.shape[0], 1))), axis=1)
            pixel_3_axis = p_concat.dot(position_to_pixel_matrix.T)[:, :3]
            pixel += [np.vstack((pixel_3_axis, pixel_3_axis[0, :]))]

        return pixel

    def update_slice_location(self, scroll, slice_plane):
        if slice_plane == 'Axial':
            self.slice_location[0] = scroll
        elif slice_plane == 'Coronal':
            self.slice_location[1] = scroll
        else:
            self.slice_location[2] = scroll


class Deformable(object):
    def __init__(self, dvf=None, origin=None, spacing=None, dimensions=None, roi_names=None, rigid_matrix=None,
                 dvf_matrix=None, registration_name=None, reference_name=None, moving_name=None, reference_sops=None,
                 moving_sops=None):
        self.reference_name = reference_name
        self.reference_sops = reference_sops
        self.moving_name = moving_name
        self.moving_sops = moving_sops
        self.roi_names = roi_names
        self.rigid_rois = dict.fromkeys(Data.roi_list)
        self.deformable_rois = dict.fromkeys(Data.roi_list)

        self.modality = None
        if dvf_matrix is not None:
            if np.allclose(dvf_matrix, np.identity(3), atol=1e-3):
                self.dvf = dvf
                self.origin = origin
                self.spacing = spacing
                self.dimensions = dimensions
            else:
                self.dvf, self.spacing, self.origin, self.dimensions = self.correct_dvf_direction(dvf, spacing, origin,
                                                                                                  dvf_matrix)

        else:
            self.dvf = dvf
            self.origin = origin
            self.spacing = spacing
            self.dimensions = dimensions

        if rigid_matrix is None:
            self.rigid_matrix = np.identity(4)
        else:
            self.rigid_matrix = rigid_matrix

        self.deformable_name = self.add_deformable(registration_name)

        self.rois = {}
        self.display = Display(self)

    def add_deformable(self, deformable_name):
        if deformable_name is None:
            if self.reference_name is None and self.moving_name is None:
                deformable_name = 'DVF_Unknown'
            else:
                deformable_name = 'DVF_' + self.reference_name + '_' + self.moving_name

            if deformable_name in Data.deformable_list:
                n = 0
                while n > -1:
                    n += 1
                    new_name = copy.deepcopy(deformable_name + '_' + str(n))
                    if new_name not in Data.deformable_list:
                        deformable_name = new_name
                        n = -100

        Data.deformable[deformable_name] = self
        Data.deformable_list += [deformable_name]

        return deformable_name

    def compute_aspect(self, slice_plane):
        if slice_plane == 'Axial':
            aspect = np.round(self.spacing[0] / self.spacing[1], 2)
        elif slice_plane == 'Coronal':
            aspect = np.round(self.spacing[0] / self.spacing[2], 2)
        else:
            aspect = np.round(self.spacing[1] / self.spacing[2], 2)

        return aspect

    def compute_biomechanical(self):
        pass

    def compute_bspline(self, modality_gradient=True, sigma=2, control_spacing=None, mesh_size=None, gradient=1e-5,
                        iterations=100, crop=5):
        ref_image = mia.Data.image[self.reference_name].create_sitk_image()
        mov_image = mia.Data.image[self.moving_name].create_sitk_image()

        euler = sitk.Euler3DTransform()
        euler.SetMatrix(self.rigid_matrix[:3, :3].flatten().tolist())
        euler.SetTranslation(self.rigid_matrix[:3, 3])

        mov_resampled = sitk.Resample(mov_image, ref_image, euler, sitk.sitkLinear, 0.0, mov_image.GetPixelID())

        ref_mask = None
        mov_mask = None
        for roi_name in self.roi_names:
            ref_roi = mia.Data.image[self.reference_name].rois[roi_name]
            mov_roi = mia.Data.image[self.moving_name].rois[roi_name]
            if ref_roi.mesh is not None or ref_roi.contour_pixel is not None:
                if mov_roi.mesh is not None or mov_roi.contour_pixel is not None:
                    if ref_mask is None:
                        ref_mask = ref_roi.compute_mask()
                    else:
                        ref_mask = ref_mask + ref_roi.compute_mask()

                        if mov_mask is None:
                            mov_mask = mov_roi.compute_mask()
                        else:
                            mov_mask = mov_mask + mov_roi.compute_mask()

        deform_itk = mia.utils.deformable.simpleitk.DeformableITK(reference_image=ref_image,
                                                                  moving_image=mov_resampled,
                                                                  reference_mask=None,
                                                                  moving_mask=None)

        if mia.Data.image[self.reference_name].modality != mia.Data.image[self.moving_name].modality and modality_gradient:
            deform_itk.cross_modality_correction()

        if ref_mask is not None and mov_mask is not None:
            deform_itk.create_sitk_image(ref_mask, ref.origin, ref.spacing, ref.matrix, mask=True)
            deform_itk.create_sitk_image(mov_mask, mov.origin, mov.spacing, mov.matrix, reference=False, mask=True)

            if sigma is not None:
                deform_itk.blur_mask(sigma=sigma)

        deform_itk.resample()
        dvf_image = deform_itk.bspline(control_spacing=control_spacing, mesh_size=mesh_size, gradient=gradient,
                                       iterations=iterations, crop=crop)

        self.origin = dvf_image.GetOrigin()
        self.spacing = dvf_image.GetSpacing()
        self.dvf = sitk.GetArrayFromImage(dvf_image)

    def compute_demons(self, method=None, modality_gradient=True, sigma=2, smooth=True, std=1, iterations=50,
                       intensity_threshold=0.001, step=2.0, crop=5):
        ref = mia.Data.image[self.reference_name]
        mov = mia.Data.image[self.moving_name]

        deform_itk = mia.utils.deformable.simpleitk.DeformableITK()
        deform_itk.create_sitk_image(ref.array, ref.origin, ref.spacing, ref.matrix)
        deform_itk.create_sitk_image(mov.array, mov.origin, mov.spacing, mov.matrix, reference=False)

        if mia.Data.image[self.reference_name].modality != mia.Data.image[self.moving_name].modality and modality_gradient:
            deform_itk.cross_modality_correction()

        ref_mask = None
        mov_mask = None
        for roi_name in self.roi_names:
            ref_roi = mia.Data.image[self.reference_name].rois[roi_name]
            mov_roi = mia.Data.image[self.moving_name].rois[roi_name]
            if ref_roi.mesh is not None or ref_roi.contour_pixel is not None:
                if mov_roi.mesh is not None or mov_roi.contour_pixel is not None:
                    if ref_mask is None:
                        ref_mask = ref_roi.compute_mask()
                    else:
                        ref_mask = ref_mask + ref_roi.compute_mask()

                        if mov_mask is None:
                            mov_mask = mov_roi.compute_mask()
                        else:
                            mov_mask = mov_mask + mov_roi.compute_mask()

        if ref_mask is not None and mov_mask is not None:
            deform_itk.create_sitk_image(ref_mask, ref.origin, ref.spacing, ref.matrix, mask=True)
            deform_itk.create_sitk_image(mov_mask, mov.origin, mov.spacing, mov.matrix, reference=False, mask=True)

            if sigma is not None:
                deform_itk.blur_mask(sigma=sigma)

        deform_itk.resample()
        if method in ['Demons', 'demons']:
            dvf_image = deform_itk.demons(smooth=smooth, std=std, iterations=iterations,
                                          intensity_threshold=intensity_threshold, crop=crop)

        elif method in ['Diffeomorphic', 'diffeomorphic']:
            dvf_image = deform_itk.diffeomorphic(smooth=smooth, std=std, iterations=iterations,
                                                 intensity_threshold=intensity_threshold, crop=crop)

        else:
            dvf_image = deform_itk.fast_demons(smooth=smooth, std=std, iterations=iterations,
                                               intensity_threshold=intensity_threshold, step=step, crop=crop)

        self.origin = dvf_image.GetOrigin()
        self.spacing = dvf_image.GetSpacing()
        self.dvf = sitk.GetArrayFromImage(dvf_image)

    @staticmethod
    def correct_dvf_direction(dvf, spacing, origin, matrix):
        """
        Rotate a DVF to a new direction, adjusting vectors and origin so that the DVF
        stays aligned in physical space (rotation about the center of the volume).
        """
        D_new = np.identity(3)
        R = D_new @ np.linalg.inv(matrix)

        center_index = (np.flip(dvf.shape)[1:] - 1) / 2.0
        center_phys = origin + matrix @ (center_index * spacing)

        Z, Y, X, _ = dvf.shape
        dvf_flat = dvf.reshape(-1, 3).T
        dvf_rotated_flat = R @ dvf_flat
        dvf_rotated = dvf_rotated_flat.T.reshape(Z, Y, X, 3)

        center_phys_new = center_phys  # center should not move
        origin_new = center_phys_new - D_new @ (center_index * spacing)

        dimensions = dvf_rotated.shape[0:3]

        return dvf_rotated, spacing, origin_new, dimensions

    def create_image(self):
        R_inv = self.rigid_matrix[:3, :3].T
        t_inv = -R_inv @ self.rigid_matrix[:3, 3]

        affine = sitk.AffineTransform(3)
        affine.SetMatrix(R_inv.flatten())
        affine.SetTranslation(t_inv)

        ref_image = Data.image[self.reference_name].create_sitk_image()
        moving_image = Data.image[self.moving_name].create_sitk_image()

        resampled_image = sitk.Resample(
            moving_image,
            size=ref_image.GetSize(),
            transform=affine,
            outputOrigin=ref_image.GetOrigin(),
            outputSpacing=ref_image.GetSpacing(),
            interpolator=sitk.sitkLinear,
            defaultPixelValue=-3001,
            outputPixelType=moving_image.GetPixelID()
        )

        dvf_initial_image = sitk.GetImageFromArray(self.dvf, isVector=True)
        dvf_initial_image.SetSpacing(self.spacing)
        dvf_initial_image.SetOrigin(self.origin)

        cropped_image = sitk.Resample(
            resampled_image,
            dvf_initial_image,
            sitk.Transform(),
            sitk.sitkLinear,
            0.0,
            resampled_image.GetPixelID()
        )

        dvf_transform = sitk.DisplacementFieldTransform(dvf_initial_image)
        dvf_resampled = sitk.ResampleImageFilter()
        dvf_resampled.SetReferenceImage(cropped_image)
        dvf_resampled.SetTransform(dvf_transform)
        dvf_resampled.SetInterpolator(sitk.sitkLinear)
        dvf_resampled.SetDefaultPixelValue(0)
        dvf_resampled.SetOutputDirection(np.identity(3).flatten().tolist())

        return dvf_resampled.Execute(resampled_image)

    def export_image(self, path=None):
        if self.moving_name is not None and path is not None:
            image = self.create_image()

            writer = vtk.vtkMetaImageWriter()
            writer.SetInputData(image)
            writer.SetFileName(path)
            writer.Write()

    def retrieve_array_plane(self, slice_plane, vector=None):
        if vector is None:
            return self.display.compute_array(slice_plane)
        elif vector == 'x':
            return self.display.compute_grid(slice_plane=slice_plane, vector=vector)
        elif vector == 'y':
            return self.display.compute_grid(slice_plane=slice_plane, vector=vector)
        elif vector == 'z':
            return self.display.compute_grid(slice_plane=slice_plane, vector=vector)

    def retrieve_grid(self, slice_plane='Axial', vector='x'):
        return self.display.compute_grid(slice_plane=slice_plane, vector=vector)

    def retrieve_slice_location(self, slice_plane):
        if slice_plane == 'Axial':
            return self.display.slice_location[0]

        elif slice_plane == 'Coronal':
            return self.display.slice_location[1]

        else:
            return self.display.slice_location[2]

    def retrieve_slice_position(self, slice_plane=None):
        pixel_to_position_matrix = self.display.compute_matrix_pixel_to_position()

        if slice_plane is None:
            location = np.asarray([self.display.slice_location[2],
                                   self.display.slice_location[1],
                                   self.display.slice_location[0], 1])
        else:
            if slice_plane == 'Axial':
                location = np.asarray([0, 0, self.display.slice_location[0], 1])
            elif slice_plane == 'Coronal':
                location = np.asarray([0, self.display.slice_location[1], 0, 1])
                print(location)
            else:
                location = np.asarray([self.display.slice_location[2], 0, 0, 1])

        return location.dot(pixel_to_position_matrix.T)[:3]

    def retrieve_scroll_max(self, slice_plane):
        if slice_plane == 'Axial':
            return self.display.scroll_max[0]

        elif slice_plane == 'Coronal':
            return self.display.scroll_max[1]

        else:
            return self.display.scroll_max[2]

    def save_deformable(self, path):
        variable_names = self.__dict__.keys()
        column_names = [name for name in variable_names if name not in ['deformable_rois',
                                                                        'display',
                                                                        'dvf',
                                                                        'rigid_rois']]

        df = pd.DataFrame(index=[0], columns=column_names)
        for name in column_names:
            df.at[0, name] = getattr(self, name)

        df.to_pickle(os.path.join(path, 'info.p'))
        np.save(os.path.join(path, 'dvf.npy'), self.dvf, allow_pickle=True)

    def update_rois(self, roi_name=None):
        for name in list(self.rois.keys()):
            if name not in Data.roi_list:
                del self.rois[name]

        for name in Data.roi_list:
            if name not in list(self.rois.keys()):
                self.rois[name] = None

        for name in Data.roi_list:
            if roi_name is None or name == roi_name:
                roi = Data.image[self.moving_name].rois[name]
                if roi.mesh is not None and roi.visible:
                    self.rigid_rois[name] = roi.mesh.transform(self.rigid_matrix, inplace=False)

                    points = self.rigid_rois[name].points
                    voxel_coords = (points - self.origin) / self.spacing
                    deformed_points = points.copy()
                    for i in range(3):
                        deformed_points[:, i] += map_coordinates(
                            self.dvf[..., i],
                            [voxel_coords[:, 2], voxel_coords[:, 1], voxel_coords[:, 0]],
                            order=1,
                            mode='nearest'
                        )

                    self.deformable_rois[name] = self.rigid_rois[name].copy()
                    self.deformable_rois.points = deformed_points




from .reader import read_3mf, read_dicoms, read_mhd

from .read import DicomReader, MhdReader, StlReader, ThreeMfReader, VtkReader
from .structure import Rigid, Deformable
from .utils import *

from .data import Data

Data()



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:

Structure:

"""

import copy

import numpy as np
import SimpleITK as sitk

import vtk
from vtkmodules.util import numpy_support

from scipy.spatial.transform import Rotation

from ..utils.rigid.icp import ICP
from ..data import Data


class Display(object):
    def __init__(self, rigid):
        self.rigid = rigid

        self.origin = None
        self.spacing = None
        self.array = None
        self.matrix = np.identity(4)

        self.slice_location = [0, 0, 0]
        self.scroll_max = None
        self.offset = {'Axial': [0, 0], 'Coronal': [0, 0], 'Sagittal': [0, 0]}
        self.misc = {}

    def compute_array_slice(self, slice_plane):
        array_slice = None
        if slice_plane == 'Axial':
            if 0 <= self.slice_location[0] < self.array.shape[0]:
                array_slice = self.array[self.slice_location[0], :, :].astype(np.double)

        elif slice_plane == 'Coronal':
            if 0 <= self.slice_location[1] < self.array.shape[1]:
                array_slice = self.array[:, self.slice_location[1], :].astype(np.double)

        else:
            if 0 <= self.slice_location[2] < self.array.shape[2]:
                array_slice = self.array[:, :, self.slice_location[2]].astype(np.double)

        return array_slice

    def compute_offset(self):
        pos = Data.image[self.rigid.reference_name].origin

        self.offset['Axial'][0] = (self.origin[0] - pos[0]) / self.spacing[0]
        self.offset['Axial'][1] = (self.origin[1] - pos[1]) / self.spacing[1]
        self.offset['Coronal'][0] = (self.origin[0] - pos[0]) / self.spacing[0]
        self.offset['Coronal'][1] = (self.origin[2] - pos[2]) / self.spacing[2]
        self.offset['Sagittal'][0] = (self.origin[1] - pos[1]) / self.spacing[1]
        self.offset['Sagittal'][1] = (self.origin[2] - pos[2]) / self.spacing[2]

    def compute_matrix_pixel_to_position(self):
        matrix = copy.deepcopy(Data.image[self.rigid.moving_name].matrix)

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * self.spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * self.spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * self.spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        return pixel_to_position_matrix

    def compute_matrix_position_to_pixel(self):
        matrix = copy.deepcopy(Data.image[self.rigid.moving_name].matrix)

        hold_matrix = np.identity(3, dtype=np.float32)
        hold_matrix[0, :] = matrix[0, :] / self.spacing[0]
        hold_matrix[1, :] = matrix[1, :] / self.spacing[1]
        hold_matrix[2, :] = matrix[2, :] / self.spacing[2]

        position_to_pixel_matrix = np.identity(4, dtype=np.float32)
        position_to_pixel_matrix[:3, :3] = hold_matrix
        position_to_pixel_matrix[:3, 3] = np.asarray(self.origin).dot(-hold_matrix.T)

        return position_to_pixel_matrix

    def compute_mesh_slice(self, roi_name=None, location=None, slice_plane=None, return_pixel=False):
        if self.rigid.rois[roi_name] is None:
            self.rigid.update_rois(roi_name=roi_name)

        if slice_plane == 'Axial':
            normal = self.matrix[:3, 2]
        elif slice_plane == 'Coronal':
            normal = self.matrix[:3, 1]
        else:
            normal = self.matrix[:3, 0]

        roi_slice = self.rigid.rois[roi_name].slice(normal=normal, origin=location)

        if return_pixel:
            if roi_slice.number_of_points > 0:
                roi_strip = roi_slice.strip(max_length=10000000)

                position = [np.asarray(c.points) for c in roi_strip.cell]
                pixels = self.convert_position_to_pixel(position=position)
                pixel_corrected = []
                for pixel in pixels:

                    if slice_plane in 'Axial':
                        pixel_reshape = pixel[:, :2]
                        pixel_corrected += [np.asarray([pixel_reshape[:, 0], pixel_reshape[:, 1]]).T]

                    elif slice_plane == 'Coronal':
                        pixel_reshape = np.column_stack((pixel[:, 0], pixel[:, 2]))
                        pixel_corrected += [pixel_reshape]

                    else:
                        pixel_reshape = pixel[:, 1:]
                        pixel_corrected += [pixel_reshape]

                return pixel_corrected

            else:
                return []

        else:
            return roi_slice

    def compute_reslice(self):
        image = self.rigid.create_image()

        self.origin = np.asarray(image.GetOrigin())
        self.spacing = image.GetSpacing()
        dimensions = image.GetDimensions()

        scalars = image.GetPointData().GetScalars()
        self.array = numpy_support.vtk_to_numpy(scalars).reshape(dimensions[2], dimensions[1], dimensions[0])

        self.compute_offset()
        self.compute_scroll_max()

    def compute_slice_location(self, position=None):
        if position is None:
            source_location = np.flip(Data.image[self.rigid.reference_name].display.slice_location)
            position = Data.image[self.rigid.reference_name].display.compute_index_positions(source_location)
        self.slice_location = np.flip(np.round((position - self.origin) / self.spacing).astype(np.int32))

    def compute_slice_origin(self, slice_plane):
        slice_origin = None
        if slice_plane == 'Axial' and 0 <= self.slice_location[0] <= self.scroll_max[0]:
            location = np.asarray([0, 0, self.slice_location[0]])
            slice_origin = self.origin + (location * self.spacing)
        elif slice_plane == 'Coronal' and 0 <= self.slice_location[1] <= self.scroll_max[1]:
            location = np.asarray([0, self.slice_location[1], 0])
            slice_origin = self.origin + (location * self.spacing)
        elif slice_plane == 'Sagittal' and 0 <= self.slice_location[2] <= self.scroll_max[2]:
            location = np.asarray([self.slice_location[2], 0, 0])
            slice_origin = self.origin + (location * self.spacing)

        return slice_origin

    def compute_scroll_max(self):
        self.scroll_max = [self.array.shape[0] - 1,
                           self.array.shape[1] - 1,
                           self.array.shape[2] - 1]

    def compute_vtk_slice(self, slice_plane):
        if self.array is None:
            self.compute_reslice()
            self.compute_scroll_max()

        self.compute_slice_location()

        slice_array = None
        slice_origin = self.compute_slice_origin(slice_plane)
        array_shape = self.array.shape
        if slice_plane == 'Axial' and 0 <= self.slice_location[0] <= self.scroll_max[0]:
            slice_array = np.zeros((1, array_shape[1], array_shape[2]))
            slice_array[0, :, :] = self.array[self.slice_location[0], :, :]

        elif slice_plane == 'Coronal' and 0 <= self.slice_location[1] <= self.scroll_max[1]:
            slice_array = np.zeros((array_shape[0], 1, array_shape[2]))
            slice_array[:, 0, :] = self.array[:, self.slice_location[1], :]

        elif slice_plane == 'Sagittal' and 0 <= self.slice_location[2] <= self.scroll_max[2]:
            slice_array = np.zeros((array_shape[0], array_shape[1], 1))
            slice_array[:, :, 0] = self.array[:, :, self.slice_location[2]]

        vtk_image = None
        if slice_array is not None:
            vtk_image = vtk.vtkImageData()
            vtk_image.SetSpacing(self.spacing)
            vtk_image.SetDirectionMatrix(1, 0, 0, 0, 1, 0, 0, 0, 1)
            vtk_image.SetDimensions(np.flip(slice_array.shape))
            vtk_image.SetOrigin(slice_origin)
            vtk_image.GetPointData().SetScalars(numpy_support.numpy_to_vtk(slice_array.flatten(order="C")))

        return vtk_image

    def convert_position_to_pixel(self, position=None):
        position_to_pixel_matrix = Data.image[self.rigid.reference_name].display.compute_matrix_position_to_pixel()

        pixel = []
        for ii, pos in enumerate(position):
            p_concat = np.concatenate((pos, np.ones((pos.shape[0], 1))), axis=1)
            pixel_3_axis = p_concat.dot(position_to_pixel_matrix.T)[:, :3]
            pixel += [np.vstack((pixel_3_axis, pixel_3_axis[0, :]))]

        return pixel

    def update_slice_location(self, scroll, slice_plane):
        # print(a)
        if slice_plane == 'Axial':
            self.slice_location[0] = scroll
        elif slice_plane == 'Coronal':
            self.slice_location[1] = scroll
        else:
            self.slice_location[2] = scroll


class Rigid(object):
    def __init__(self, reference_name, moving_name, rigid_name=None, roi_names=None, reference_sops=None,
                 moving_sops=None, reference_matrix=None, matrix=None, combo_matrix=None, combo_name=None):
        self.reference_name = reference_name
        self.moving_name = moving_name
        self.combo_name = combo_name
        self.rois = dict.fromkeys(Data.roi_list)

        if roi_names is None:
            self.roi_names = ['Unknown']
        else:
            self.roi_names = roi_names

        if reference_matrix is None:
            self.reference_matrix = np.identity(4)
        else:
            self.reference_matrix = reference_matrix

        if matrix is None:
            self.matrix = np.identity(4)
        else:
            self.matrix = matrix

        if combo_matrix is None:
            self.combo_matrix = np.identity(4)
        else:
            self.combo_matrix = combo_matrix

        self.slices = {'reference': ['All'], 'moving': ['All'], 'reference_sops': reference_sops,
                       'moving_sops': moving_sops}
        self.visual = {'reference': None, 'moving': None, 'opacity': 0.5, 'multicolor': None}

        self.misc = {}
        self.rotation_center = np.asarray([0, 0, 0])
        self.rigid_name = self.add_rigid(rigid_name)

        self.display = Display(self)

    def add_rigid(self, rigid_name):
        if rigid_name is None:
            if np.array_equal(self.combo_matrix, np.identity(4)):
                rigid_name = self.reference_name + '_' + self.moving_name
            else:
                rigid_name = self.reference_name + '_' + self.moving_name + '_combo'

            if rigid_name in Data.rigid_list:
                n = 0
                while n > -1:
                    n += 1
                    new_name = copy.deepcopy(rigid_name + '_' + str(n))
                    if new_name not in Data.rigid_list:
                        rigid_name = new_name
                        n = -100

        Data.rigid[rigid_name] = self
        Data.rigid_list += [rigid_name]

        return rigid_name

    def compute_aspect(self, slice_plane):
        if slice_plane == 'Axial':
            aspect = np.round(self.display.spacing[0] / self.display.spacing[1], 2)
        elif slice_plane == 'Coronal':
            aspect = np.round(self.display.spacing[0] / self.display.spacing[2], 2)
        else:
            aspect = np.round(self.display.spacing[1] / self.display.spacing[2], 2)

        return aspect

    def compute_icp_vtk(self, source_mesh, target_mesh, distance=1e-5, iterations=1000, landmarks=None,
                        com_matching=True, inverse=True, center=None):
        target_mesh.transform(self.matrix @ self.combo_matrix, inplace=True)

        icp = ICP(source_mesh, target_mesh)
        icp.compute_vtk(distance=distance, iterations=iterations, landmarks=landmarks, com_matching=com_matching,
                        inverse=inverse)

        if center == 'image':
            R_icp = np.asarray(icp.get_matrix(), dtype=float)
            old_center = np.array([0, 0, 0], dtype=float)
            new_center = np.array(Data.image[self.moving_name].compute_center(), dtype=float)

            T_neg = np.eye(4)
            T_neg[:3, 3] = -new_center
            T_pos = np.eye(4)
            T_pos[:3, 3] = new_center

            extra_rotation = np.eye(4)
            old_center_h = np.hstack([old_center, 1])
            new_center_h = np.hstack([new_center, 1])

            R_total = extra_rotation @ R_icp
            transformed_old_center = R_total @ old_center_h
            transformed_new_center = R_total @ new_center_h

            correction = (old_center_h - transformed_old_center) - (new_center_h - transformed_new_center)
            T_corr = np.eye(4)
            T_corr[:3, 3] = correction[:3]
            self.matrix = T_pos @ extra_rotation @ R_icp @ T_neg @ T_corr

        else:
            self.matrix = icp.get_matrix()

        self.update_rois()

    def compute_o3d(self, source_mesh, target_mesh, distance=10, iterations=1000, rmse=1e-7, fitness=1e-7,
                    method='point', com_matching=True, inverse=True, center=None):
        target_mesh.transform(self.matrix @ self.combo_matrix, inplace=True)

        icp = ICP(source_mesh, target_mesh)
        icp.compute_o3d(distance=distance, iterations=iterations, rmse=rmse, fitness=fitness, method=method,
                        com_matching=com_matching, inverse=inverse)

        if center == 'image':
            R_icp = np.asarray(icp.get_matrix(), dtype=float)
            old_center = np.array([0, 0, 0], dtype=float)
            new_center = np.array(Data.image[self.moving_name].compute_center(), dtype=float)

            T_neg = np.eye(4)
            T_neg[:3, 3] = -new_center
            T_pos = np.eye(4)
            T_pos[:3, 3] = new_center

            extra_rotation = np.eye(4)
            old_center_h = np.hstack([old_center, 1])
            new_center_h = np.hstack([new_center, 1])

            R_total = extra_rotation @ R_icp
            transformed_old_center = R_total @ old_center_h
            transformed_new_center = R_total @ new_center_h

            correction = (old_center_h - transformed_old_center) - (new_center_h - transformed_new_center)
            T_corr = np.eye(4)
            T_corr[:3, 3] = correction[:3]
            self.matrix = T_pos @ extra_rotation @ R_icp @ T_neg @ T_corr

        else:
            self.matrix = icp.get_matrix()

        self.update_rois()

    def copy_roi(self, roi_name=None, inverse=False):
        if roi_name in list(self.rois.keys()):
            reference_roi = Data.image[self.reference_name].rois[roi_name]
            moving_roving = Data.image[self.moving_name].rois[roi_name]
            if inverse and self.rois[roi_name] is not None:
                reference_roi.mesh = self.rois[roi_name].transform(self.matrix @ self.combo_matrix, inplace=False)
            elif reference_roi.mesh is not None:
                moving_roving.mesh = reference_roi.mesh.transform(np.linalg.inv(self.matrix @ self.combo_matrix),
                                                                  inplace=False)
                self.update_rois(roi_name=roi_name)

    def create_image(self):
        name = self.moving_name
        matrix_reshape = Data.image[name].matrix.reshape(1, 9)[0]
        vtk_image = vtk.vtkImageData()
        vtk_image.SetSpacing(Data.image[name].spacing)
        vtk_image.SetDirectionMatrix(matrix_reshape)
        vtk_image.SetDimensions(np.flip(Data.image[name].array.shape))
        vtk_image.SetOrigin(Data.image[name].origin)
        vtk_image.GetPointData().SetScalars(numpy_support.numpy_to_vtk(Data.image[name].array.flatten(order="C")))

        matrix = self.matrix @ self.combo_matrix
        vtk_matrix = vtk.vtkMatrix4x4()
        for i in range(4):
            for j in range(4):
                vtk_matrix.SetElement(i, j, matrix[i, j])

        transform = vtk.vtkTransform()
        transform.SetMatrix(vtk_matrix)
        transform.Inverse()

        vtk_reslice = vtk.vtkImageReslice()
        vtk_reslice.SetInputData(vtk_image)
        vtk_reslice.SetResliceTransform(transform)
        vtk_reslice.SetInterpolationModeToLinear()
        vtk_reslice.SetOutputSpacing(Data.image[self.reference_name].spacing)
        vtk_reslice.SetOutputDirection(1, 0, 0, 0, 1, 0, 0, 0, 1)
        vtk_reslice.AutoCropOutputOn()
        vtk_reslice.SetBackgroundLevel(-3001)
        vtk_reslice.Update()

        return vtk_reslice.GetOutput()

    def export_image(self, path=None):
        if self.moving_name is not None and path is not None:
            image = self.create_image()

            writer = vtk.vtkMetaImageWriter()
            writer.SetInputData(image)
            writer.SetFileName(path)
            writer.Write()

    def pre_alignment(self, superior=False, center=False, origin=False):
        if superior:
            pass
        elif center:
            pass
        elif origin:
            self.matrix[:3, 3] = Data.image[self.reference_name].origin - Data.image[self.moving_name].origin

    def retrieve_angles(self, order='ZXY'):
        rotation = Rotation.from_matrix(self.matrix[:3, :3])
        return rotation.as_euler(order, degrees=True)

    def retrieve_array_plane(self, slice_plane, solo=None, position=None):
        if self.display.array is None:
            self.display.compute_reslice()
            self.display.compute_scroll_max()

        if solo is None:
            self.display.compute_slice_location(position=position)

        return self.display.compute_array_slice(slice_plane=slice_plane)

    def retrieve_offset(self, slice_plane):
        return self.display.offset[slice_plane]

    def retrieve_slice_location(self, slice_plane):
        if slice_plane == 'Axial':
            return self.display.slice_location[0]

        elif slice_plane == 'Coronal':
            return self.display.slice_location[1]

        else:
            return self.display.slice_location[2]

    def retrieve_slice_position(self, slice_plane=None):
        pixel_to_position_matrix = self.display.compute_matrix_pixel_to_position()

        if slice_plane is None:
            location = np.asarray([self.display.slice_location[2],
                                   self.display.slice_location[1],
                                   self.display.slice_location[0], 1])
        else:
            if slice_plane == 'Axial':
                location = np.asarray([0, 0, self.display.slice_location[0], 1])
            elif slice_plane == 'Coronal':
                location = np.asarray([0, self.display.slice_location[1], 0, 1])
                print(location)
            else:
                location = np.asarray([self.display.slice_location[2], 0, 0, 1])

        return location.dot(pixel_to_position_matrix.T)[:3]

    def retrieve_scroll_max(self, slice_plane):
        if slice_plane == 'Axial':
            return self.display.scroll_max[0]

        elif slice_plane == 'Coronal':
            return self.display.scroll_max[1]

        else:
            return self.display.scroll_max[2]

    def retrieve_translation(self):
        return self.matrix[:3, 3]

    def retrieve_vtk_slice(self, slice_plane):
        return self.display.compute_vtk_slice(slice_plane)

    def save_rigid(self, path):
        variable_names = self.__dict__.keys()
        column_names = [name for name in variable_names if name not in ['rois', 'pois', 'display']]

        df = pd.DataFrame(index=[0], columns=column_names)
        for name in column_names:
            df.at[0, name] = getattr(self, name)

        df.to_pickle(os.path.join(path, 'info.p'))

    def update_rotation(self, r_x=0, r_y=0, r_z=0):
        new_matrix = np.identity(4)
        if r_x:
            radians = np.deg2rad(r_x)
            new_matrix[:3, :3] = Rotation.from_euler('x', radians).as_matrix()
        if r_y:
            radians = np.deg2rad(r_y)
            new_matrix[:3, :3] = Rotation.from_euler('y', radians).as_matrix()
        if r_z:
            radians = np.deg2rad(r_z)
            new_matrix[:3, :3] = Rotation.from_euler('z', radians).as_matrix()

        self.matrix = new_matrix @ self.matrix
        self.display.compute_reslice()
        self.display.compute_scroll_max()
        self.update_rois()

    def update_translation(self, t_x=0, t_y=0, t_z=0):
        new_matrix = np.identity(4)
        new_matrix[0, 3] = t_x
        new_matrix[1, 3] = t_y
        new_matrix[2, 3] = t_z

        self.matrix = new_matrix @ self.matrix
        self.display.origin[0] = self.display.origin[0] + t_x
        self.display.origin[1] = self.display.origin[1] + t_y
        self.display.origin[2] = self.display.origin[2] + t_z

        self.display.compute_offset()
        self.display.compute_scroll_max()
        self.update_rois()

    def update_rois(self, roi_name=None):
        for name in list(self.rois.keys()):
            if name not in Data.roi_list:
                del self.rois[name]

        for name in Data.roi_list:
            if name not in list(self.rois.keys()):
                self.rois[name] = None

        for name in Data.roi_list:
            if roi_name is None or name == roi_name:
                roi = Data.image[self.moving_name].rois[name]
                if roi.mesh is not None and roi.visible:
                    self.rois[roi_name] = roi.mesh.transform(self.matrix @ self.combo_matrix, inplace=False)



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:
    Under construction

Structure:

"""

import cv2
import vtk
import pyacvd
import pymeshfix
import numpy as np
import pyvista as pv

from scipy.spatial import distance


class Refinement(object):
    """
    Provides tools for refining and processing surface meshes, including smoothing, clustering, decimation, and advanced
    face splitting for mesh quality improvement.
    """
    def __init__(self, mesh):
        self.mesh = mesh

        self.correct_faces = None

        self.points = np.asarray(self.mesh.points)
        self.face = mesh.faces.reshape(int(len(mesh.faces) / 4), 4)[:, 1:]
        self.face_centers = np.asarray(mesh.cell_centers().points)
        self.face_lines_sort = np.sort(np.vstack([[[ff[0], ff[1]], [ff[0], ff[2]], [ff[1], ff[2]]] for ff in self.face]), axis=1)
        self.face_lines = np.unique(self.face_lines_sort, axis=0)

    def smooth(self, iterations=20, angle=60, passband=0.001):
        """
        Smooths the mesh geometry using a windowed sinc filter.

        Parameters
        ----------
        iterations : int
            Number of smoothing iterations.
        angle : float
            Feature angle in degrees for smoothing sharp edges.
        passband : float
            Passband for the filter (controls smoothness).

        Returns
        -------
        mesh : pyvista.PolyData
            Smoothed mesh.
        """
        smoother = vtk.vtkWindowedSincPolyDataFilter()
        smoother.SetInputData(self.mesh)
        smoother.SetNumberOfIterations(iterations)
        smoother.BoundarySmoothingOff()
        smoother.FeatureEdgeSmoothingOff()
        smoother.SetFeatureAngle(angle)
        smoother.SetPassBand(passband)
        smoother.NonManifoldSmoothingOn()
        smoother.NormalizeCoordinatesOff()
        # smoother.SetRelaxationFactor(0.1)
        smoother.Update()
        self.mesh = pv.PolyData(smoother.GetOutput())

        return self.mesh

    def cluster(self, points=None):
        """
        Reduces the number of points on the mesh using clustering.

        Parameters
        ----------
        points : int, optional
            Number of target points. If None, computed automatically.

        Returns
        -------
        mesh : pyvista.PolyData
            Mesh after clustering.
        """
        if points is None:
            points = self.compute_points()
        clus = pyacvd.Clustering(self.mesh)
        clus.cluster(points)
        self.mesh = clus.create_mesh()

        return self.mesh

    def decimate(self, percent=None):
        """
        Decimates (reduces) mesh complexity by removing points and faces.

        Parameters
        ----------
        percent : float, optional
            Fraction of the mesh to remove. If None, computed automatically.

        Returns
        -------
        mesh : pyvista.PolyData
            Decimated mesh.
        """
        if percent is None:
            percent = self.compute_point_percentage()

        self.mesh.decimate(percent)

        return self.mesh

    def compute_points(self):
        """
        Computes a target number of points based on the mesh size.

        Returns
        -------
        int
            Approximate number of points for clustering/decimation.
        """
        return np.round(10 * np.sqrt(self.mesh.number_of_points))

    def compute_point_percentage(self):
        """
                Computes the fraction of points to remove for decimation.

                Returns
                -------
                float
                    Percentage of points to remove.
                """
        points = self.compute_points()

        return 1 - (points / self.mesh.number_of_points)

    def tri_split(self):
        """
        Splits selected faces into smaller triangles by adding face centroids.

        Returns
        -------
        pyvista.PolyData
            Mesh with refined triangular faces.
        """
        self.find_face_correction()

        base_faces = [f for ii, f in enumerate(self.face) if ii not in self.correct_faces]
        base_length = len(self.points)
        new_points = [self.face_centers[ii] for ii in self.correct_faces]
        total_points = np.concatenate((self.points, new_points))

        new_faces = []
        for ii, f in enumerate(self.correct_faces):
            hold_face = self.face[f]
            new_faces += [[hold_face[0], hold_face[1], base_length + ii]]
            new_faces += [[hold_face[1], hold_face[2], base_length + ii]]
            new_faces += [[hold_face[0], hold_face[2], base_length + ii]]

        total_faces = np.concatenate((base_faces, np.asarray(new_faces)))
        face_syntax = [[3, t[0], t[1], t[2]] for t in total_faces]

        return pv.PolyData(total_points, face_syntax)

    def advanced_split(self):
        """
        Advanced face splitting using midpoints of edges and face corrections. Intended for complex mesh refinement
        where tri_split is insufficient.
        """
        self.find_face_correction()

        midpoint, midline = self.compute_midpoints()
        all_points = np.concatenate(self.points, midpoint)

        face_line_help = np.asarray(
            [np.where((self.face_lines_sort[:, 0] == m[0]) & (self.face_lines_sort[:, 1] == m[1]))[0] for m in midline])
        face_help = [[np.sort(self.face[int(f[0] / 3)]), np.sort(self.face[int(f[1] / 3)])] for f in face_line_help]
        face_help_stack = np.vstack(face_help)
        face_unique = np.unique(face_help_stack, axis=0)

        for f in face_unique:
            mid_1 = np.where((midline[:, 0] == f[0]) & (midline[:, 1] == f[1]))[0]
            mid_2 = np.where((midline[:, 0] == f[1]) & (midline[:, 1] == f[2]))[0]
            mid_3 = np.where((midline[:, 0] == f[0]) & (midline[:, 1] == f[2]))[0]

            point_1 = self.points[f[0]]
            point_2 = self.points[f[1]]
            point_3 = self.points[f[2]]
            center = np.mean(np.asarray([point_1, point_2, point_3]), axis=0)
            if len(mid_1) + len(mid_2) + len(mid_3) == 3:
                print(1)

    def find_face_correction(self):
        """
        Identifies faces that require splitting based on distance between face centers.
        """
        dist = distance.cdist(self.face_centers, self.face_centers, metric='euclidean')
        dist_sort = np.sort(dist, axis=1)
        dist_sum = np.sum(dist_sort[:, :6], axis=1)
        dist_sum_sort = np.argsort(dist_sum)
        self.correct_faces = dist_sum_sort[:int(len(self.points) / 4)]

    def compute_midpoints(self):
        """
        Computes midpoints of mesh edges for advanced splitting.

        Returns
        -------
        midpoint_unique : np.ndarray
            Unique midpoint coordinates.
        midline_unique : np.ndarray
            Corresponding edge indices for midpoints.
        """
        line_midpoints = np.vstack([[(self.points[line[0]] + self.points[line[1]]) / 2] for line in self.face_lines])

        midpoint_list = []
        midline_list = []
        for ii, c in enumerate(correct_faces):
            f = face[c]
            midpoint_1 = (self.points[f[0]] + self.points[f[1]]) / 2
            midpoint_2 = (self.points[f[1]] + self.points[f[2]]) / 2
            midpoint_3 = (self.points[f[2]] + self.points[f[0]]) / 2

            dist_1 = np.linalg.norm(midpoint_1 - self.points[f[2]])
            dist_2 = np.linalg.norm(midpoint_2 - self.points[f[0]])
            dist_3 = np.linalg.norm(midpoint_3 - self.points[f[1]])

            if dist_1 < dist_2 and dist_1 < dist_3:
                f_sort = np.sort([f[0], f[1]])
                midpoint_list += [
                    line_midpoints[np.where((face_lines[:, 0] == f_sort[0]) & (face_lines[:, 1] == f_sort[1]))[0][0]]]
                midline_list += [f_sort]
            elif dist_2 < dist_1 and dist_2 < dist_3:
                f_sort = np.sort([f[1], f[2]])
                midpoint_list += [
                    line_midpoints[np.where((face_lines[:, 0] == f_sort[0]) & (face_lines[:, 1] == f_sort[1]))[0][0]]]
                midline_list += [f_sort]
            else:
                f_sort = np.sort([f[2], f[0]])
                midpoint_list += [
                    line_midpoints[np.where((face_lines[:, 0] == f_sort[0]) & (face_lines[:, 1] == f_sort[1]))[0][0]]]
                midline_list += [f_sort]

        midpoint_unique, indices = np.unique(midpoint_list, axis=0, return_index=True)
        midline_unique = np.asarray(midline_list)[indices]

        return midpoint_unique, midline_unique


def clean_mesh(mesh):
    """
    Cleans a mesh by removing holes, duplicate vertices, and other inconsistencies.

    Parameters
    ----------
    mesh : pyvista.PolyData
        Input mesh to be cleaned.

    Returns
    -------
    corrected_mesh : pyvista.PolyData
        Cleaned mesh.
    """
    mesh = mesh.copy()

    meshfix = pymeshfix.PyTMesh()
    meshfix.load_array(mesh.points, mesh.faces.reshape((-1, 4))[:, 1:])
    meshfix.clean()

    verts, faces = meshfix.return_arrays()
    new_faces = np.insert(faces, 0, 3, axis=1)
    corrected_mesh = pv.PolyData(verts, new_faces)

    return corrected_mesh


def expansion(mesh, dist):
    """
    Expands a mesh along its vertex normals by a given distance and cleans the resulting mesh.

    Parameters
    ----------
    mesh : pyvista.PolyData
        Input mesh to be expanded.
    dist : float
        Distance to expand along vertex normals.

    Returns
    -------
    corrected_mesh : pyvista.PolyData
        Expanded and cleaned mesh.
    """
    mesh = mesh.copy()
    mesh.points += mesh.point_normals * dist

    meshfix = pymeshfix.PyTMesh()
    meshfix.load_array(mesh.points, mesh.faces.reshape((-1, 4))[:, 1:])
    meshfix.clean()

    verts, faces = meshfix.return_arrays()
    new_faces = np.insert(faces, 0, 3, axis=1)
    corrected_mesh = pv.PolyData(verts, new_faces)

    return corrected_mesh


def surface_boundary(source_meshes, target_meshes, points, matrix=None):
    """
    Aligns source and target meshes using point clustering to ensure that both meshes
    have the same number of points along corresponding boundaries.

    Parameters
    ----------
    source_meshes : list of pyvista.PolyData
        List of source meshes.
    target_meshes : list of pyvista.PolyData
        List of target meshes.
    points : list of arrays
        Reference points for clustering each mesh.
    matrix : np.array, optional
        Transformation matrix to apply to target meshes.

    Returns
    -------
    new_sources, new_targets : lists of pyvista.PolyData
        Lists of aligned and transformed meshes.
    """
    if matrix is None:
        matrix = np.identity(4)

    new_sources = []
    new_targets = []
    for ii, s in enumerate(source_meshes):
        n = -1
        while n >= -1:
            n += 1

            refine = Refinement(s)
            hold_s = refine.cluster(points=points[ii] + n)

            refine = Refinement(target_meshes[ii])
            hold_t = refine.cluster(points=points[ii] + n)

            if len(hold_s.points) == len(hold_t.points):
                n = -100

                new_sources += [hold_s]
                new_targets += [hold_t.transform(matrix, inplace=True)]

    return new_sources, new_targets


def only_main_component(mesh):
    """
    Extracts only the largest connected component of a mesh.

    Parameters
    ----------
    mesh : pyvista.PolyData
        Input mesh which may contain multiple disconnected components.

    Returns
    -------
    new_mesh : pyvista.PolyData
        The largest connected component, returned as a surface mesh.
    """
    multi_block = mesh.split_bodies()

    if len(multi_block) == 1:
        return mesh

    else:
        total_points = [len(m.points) for m in multi_block]
        idx = np.argmax(total_points)
        new_mesh = multi_block[int(idx)]

        return new_mesh.extract_surface()



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:

Structure:

"""

import os
import copy

import numpy as np
import pandas as pd
import pyvista as pv
import SimpleITK as sitk

from scipy.spatial.transform import Rotation

import vtk
from vtkmodules.util import numpy_support

from ..utils.image.threshold import external
from ..utils.roi.contour import contours_from_mask

from .poi import Poi
from .roi import Roi


class Display(object):
    def __init__(self, image):
        self.image = image

        self.matrix = copy.deepcopy(self.image.matrix)
        self.spacing = copy.deepcopy(self.image.spacing)
        self.origin = copy.deepcopy(self.image.origin)

        self.slice_location = self.image.compute_center(position=False, zyx=True)
        self.scroll_max = [self.image.dimensions[0] - 1,
                           self.image.dimensions[1] - 1,
                           self.image.dimensions[2] - 1]
        self.secondary_array = None
        self.misc = {}

    def compute_matrix_pixel_to_position(self):
        matrix = copy.deepcopy(self.matrix)
        spacing = self.spacing

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        return pixel_to_position_matrix

    def compute_matrix_position_to_pixel(self):
        matrix = copy.deepcopy(self.matrix)
        spacing = self.spacing

        hold_matrix = np.identity(3, dtype=np.float32)
        hold_matrix[0, :] = matrix[0, :] / spacing[0]
        hold_matrix[1, :] = matrix[1, :] / spacing[1]
        hold_matrix[2, :] = matrix[2, :] / spacing[2]

        position_to_pixel_matrix = np.identity(4, dtype=np.float32)
        position_to_pixel_matrix[:3, :3] = hold_matrix
        position_to_pixel_matrix[:3, 3] = np.asarray(self.origin).dot(-hold_matrix.T)

        return position_to_pixel_matrix

    def compute_array(self, slice_plane):
        if self.secondary_array is None:
            if slice_plane == 'Axial':
                array = self.image.array[self.slice_location[0], :, :]
            elif slice_plane == 'Coronal':
                array = self.image.array[:, self.slice_location[1], :]
            else:
                array = self.image.array[:, :, self.slice_location[2]]
        else:
            if slice_plane == 'Axial':
                array = self.secondary_array[self.slice_location[0], :, :]
            elif slice_plane == 'Coronal':
                array = self.secondary_array[:, self.slice_location[1], :]
            else:
                array = self.secondary_array[:, :, self.slice_location[2]]

        return array.astype(np.float32)

    def compute_index_positions(self, xyz):
        pixel_to_position_matrix = self.compute_matrix_pixel_to_position()
        location = np.asarray([xyz[0], xyz[1], xyz[2], 1])

        return location.dot(pixel_to_position_matrix.T)[:3]

    def compute_offaxis_array(self):
        loc = np.flip(self.slice_location)
        base_position_matrix = self.compute_matrix_pixel_to_position()
        slice_position = np.asarray([loc[0], loc[1], loc[2], 1]).dot(base_position_matrix.T)[:3]

        matrix_reshape = self.image.matrix.reshape(1, 9)[0]
        vtk_image = vtk.vtkImageData()
        vtk_image.SetSpacing(self.image.spacing)
        vtk_image.SetDirectionMatrix(matrix_reshape)
        vtk_image.SetDimensions(np.flip(self.image.array.shape))
        vtk_image.SetOrigin(self.image.origin)
        vtk_image.GetPointData().SetScalars(numpy_support.numpy_to_vtk(self.image.array.flatten(order="C")))

        matrix = vtk.vtkMatrix4x4()
        for i in range(3):
            for j in range(3):
                matrix.SetElement(i, j, self.matrix[i, j])

        transform = vtk.vtkTransform()
        transform.SetMatrix(matrix)
        transform.Inverse()

        vtk_reslice = vtk.vtkImageReslice()
        vtk_reslice.SetInputData(vtk_image)
        vtk_reslice.SetResliceTransform(transform)
        vtk_reslice.SetInterpolationModeToLinear()
        vtk_reslice.SetOutputSpacing(self.image.spacing)
        vtk_reslice.AutoCropOutputOn()
        vtk_reslice.SetBackgroundLevel(-3001)
        vtk_reslice.Update()

        reslice_data = vtk_reslice.GetOutput()
        new_origin = reslice_data.GetOrigin()
        self.origin = transform.TransformPoint(new_origin)
        dimensions = reslice_data.GetDimensions()

        position_to_pixel_matrix = self.compute_matrix_position_to_pixel()
        location = np.asarray([slice_position[0], slice_position[1], slice_position[2], 1])
        self.slice_location = list(np.flip(np.round(location.dot(position_to_pixel_matrix.T)[:3])).astype(np.int32))
        self.scroll_max = [dimensions[2] - 1, dimensions[1] - 1, dimensions[0] - 1]
        if self.slice_location[0] > dimensions[2] - 1:
            self.slice_location[0] = dimensions[2] - 1
        if self.slice_location[1] > dimensions[1] - 1:
            self.slice_location[1] = dimensions[1] - 1
        if self.slice_location[2] > dimensions[0] - 1:
            self.slice_location[2] = dimensions[0] - 1

        scalars = reslice_data.GetPointData().GetScalars()
        self.secondary_array = numpy_support.vtk_to_numpy(scalars).reshape(dimensions[2], dimensions[1], dimensions[0])

    def compute_scroll_max(self):
        if self.secondary_array is not None:
            self.scroll_max = [self.secondary_array.shape[0] - 1,
                               self.secondary_array.shape[1] - 1,
                               self.secondary_array.shape[2] - 1]
        else:
            self.scroll_max = [self.image.dimensions[0] - 1,
                               self.image.dimensions[1] - 1,
                               self.image.dimensions[2] - 1]

    def compute_vtk_slice(self, slice_plane):
        matrix_reshape = np.linalg.inv(self.matrix).reshape(1, 9)[0]
        pixel_to_position_matrix = self.compute_matrix_pixel_to_position()
        if slice_plane == 'Axial':
            location = np.asarray([0, 0, self.slice_location[0], 1])
            if self.secondary_array is None:
                array_slice = self.image.array[self.slice_location[0], :, :]
            else:
                array_slice = self.secondary_array[self.slice_location[0], :, :]
            array_shape = array_slice.shape
            dim = [array_shape[1], array_shape[0], 1]
        elif slice_plane == 'Coronal':
            location = np.asarray([0, self.slice_location[1], 0, 1])
            if self.secondary_array is None:
                array_slice = self.image.array[:, self.slice_location[1], :]
            else:
                array_slice = self.secondary_array[:, self.slice_location[1], :]
            array_shape = array_slice.shape
            dim = [array_shape[1], 1, array_shape[0]]
        else:
            location = np.asarray([self.slice_location[2], 0, 0, 1])
            if self.secondary_array is None:
                array_slice = self.image.array[:, :, self.slice_location[2]]
            else:
                array_slice = self.secondary_array[:, :, self.slice_location[2]]
            array_shape = array_slice.shape
            dim = [1, array_shape[1], array_shape[0]]

        slice_origin = location.dot(pixel_to_position_matrix.T)[:3]

        vtk_test = vtk.vtkImageData()
        vtk_test.SetSpacing(self.image.spacing)
        vtk_test.SetDirectionMatrix(matrix_reshape)
        vtk_test.SetDimensions(dim)
        vtk_test.SetOrigin(slice_origin)
        vtk_test.GetPointData().SetScalars(numpy_support.numpy_to_vtk(array_slice.flatten(order="C")))

        return vtk_test

    def update_slice_location(self, scroll, slice_plane):
        if slice_plane == 'Axial':
            self.slice_location[0] = scroll
        elif slice_plane == 'Coronal':
            self.slice_location[1] = scroll
        else:
            self.slice_location[2] = scroll


class Image(object):
    def __init__(self, image):
        self.rois = {}
        self.pois = {}

        self.tags = image.image_set
        self.array = image.array

        self.image_name = image.image_name
        self.modality = image.modality

        self.patient_name = self.get_patient_name()
        self.mrn = self.get_mrn()
        self.birthdate = self.get_birthdate()
        self.date = self.get_date()
        self.time = self.get_time()
        self.series_uid = self.get_series_uid()
        self.acq_number = self.get_acq_number()
        self.frame_ref = self.get_frame_ref()
        self.window = self.get_window()

        self.filepaths = image.filepaths
        self.sops = image.sops

        self.plane = image.plane
        self.spacing = image.spacing
        self.dimensions = image.dimensions
        self.orientation = image.orientation
        self.origin = image.origin
        self.matrix = image.image_matrix

        self.unverified = image.unverified
        self.skipped_slice = image.skipped_slice
        self.sections = image.sections
        self.rgb = image.rgb

        self.camera_position = None
        self.misc = {}

        self.display = Display(self)

    def input_rtstruct(self, rtstruct):
        for ii, roi_name in enumerate(rtstruct.roi_names):
            if roi_name not in list(self.rois.keys()):
                self.rois[roi_name] = Roi(self, position=rtstruct.contours[ii], name=roi_name,
                                          color=rtstruct.roi_colors[ii], visible=False, filepaths=rtstruct.filepaths)

        for ii, poi_name in enumerate(rtstruct.poi_names):
            if poi_name not in list(self.pois.keys()):
                self.pois[poi_name] = Poi(self, position=rtstruct.points[ii], name=poi_name,
                                          color=rtstruct.poi_colors[ii], visible=False, filepaths=rtstruct.filepaths)

    def add_roi(self, roi_name=None, color=None, visible=False, path=None, contour=None, plane='Axial'):
        self.rois[roi_name] = Roi(self, position=contour, name=roi_name, color=color, visible=visible, filepaths=path,
                                  plane=plane)

    def add_poi(self, poi_name=None, color=None, visible=False, path=None, point=None):
        self.pois[poi_name] = Poi(self, position=point, name=poi_name, color=color, visible=visible, filepaths=path)

    def create_roi(self, name=None, color=None, visible=False, filepath=None):
        self.rois[name] = Roi(self, name=name, color=color, visible=visible, filepaths=filepath)

    def get_patient_name(self):
        if 'PatientName' in self.tags[0]:
            return str(self.tags[0].PatientName).split('^')[:3]
        else:
            return 'missing'

    def get_mrn(self):
        if 'PatientID' in self.tags[0]:
            return str(self.tags[0].PatientID)
        else:
            return 'missing'

    def get_birthdate(self):
        if 'PatientBirthDate' in self.tags[0]:
            return str(self.tags[0].PatientBirthDate)
        else:
            return ''

    def get_date(self):
        if 'SeriesDate' in self.tags[0]:
            return self.tags[0].SeriesDate
        elif 'ContentDate' in self.tags[0]:
            return self.tags[0].ContentDate
        elif 'AcquisitionDate' in self.tags[0]:
            return self.tags[0].AcquisitionDate
        elif 'StudyDate' in self.tags[0]:
            return self.tags[0].StudyDate
        else:
            return '00000'

    def get_time(self):
        if 'SeriesTime' in self.tags[0]:
            return self.tags[0].SeriesTime
        elif 'ContentTime' in self.tags[0]:
            return self.tags[0].ContentTime
        elif 'AcquisitionTime' in self.tags[0]:
            return self.tags[0].AcquisitionTime
        elif 'StudyTime' in self.tags[0]:
            return self.tags[0].StudyTime
        else:
            return '00000'

    def get_study_uid(self):
        if 'StudyInstanceUID' in self.tags[0]:
            return self.tags[0].StudyInstanceUID
        else:
            return '00000.00000'

    def get_series_uid(self):
        if 'SeriesInstanceUID' in self.tags[0]:
            return self.tags[0].SeriesInstanceUID
        else:
            return '00000.00000'

    def get_acq_number(self):
        if 'AcquisitionNumber' in self.tags[0]:
            return self.tags[0].AcquisitionNumber
        else:
            return '1'

    def get_frame_ref(self):
        if 'FrameOfReferenceUID' in self.tags[0]:
            return self.tags[0].FrameOfReferenceUID
        else:
            return '00000.00000'

    def get_window(self):
        if (0x0028, 0x1050) in self.tags[0] and (0x0028, 0x1051) in self.tags[0]:
            center = self.tags[0].WindowCenter
            width = self.tags[0].WindowWidth

            if not isinstance(center, float):
                center = center[0]

            if not isinstance(width, float):
                width = width[0]

            return [int(center) - int(np.round(width / 2)), int(center) + int(np.round(width / 2))]

        elif self.array is not None:
            return [np.min(self.array), np.max(self.array)]

        else:
            return [0, 1]

    def get_specific_tag(self, tag):
        if tag in self.tags[0]:
            return self.tags[0][tag]
        else:
            return None

    def get_specific_tag_on_all_files(self, tag):
        if tag in self.tags[0]:
            return [t[tag] for t in self.tags]
        else:
            return None

    def save_image(self, path, rois=True, pois=True):
        variable_names = self.__dict__.keys()
        column_names = [name for name in variable_names if name not in ['rois', 'pois', 'tags', 'array', 'display']]

        df = pd.DataFrame(index=[0], columns=column_names)
        for name in column_names:
            df.at[0, name] = getattr(self, name)

        df.to_pickle(os.path.join(path, 'info.p'))
        np.save(os.path.join(path, 'tags.npy'), self.tags, allow_pickle=True)
        np.save(os.path.join(path, 'array.npy'), self.array, allow_pickle=True)

        if rois:
            self.save_rois(path, create_main_folder=True)

        if pois:
            self.save_pois(path, create_main_folder=True)

    def save_rois(self, path, create_main_folder=False):
        if create_main_folder:
            path = os.path.join(path, 'ROIs')
            os.mkdir(path)

        for name in list(self.rois.keys()):
            roi_path = os.path.join(os.path.join(path, name))
            os.mkdir(roi_path)

            np.save(os.path.join(roi_path, 'name.npy'), self.rois[name].name, allow_pickle=True)
            np.save(os.path.join(roi_path, 'visible.npy'), self.rois[name].visible, allow_pickle=True)
            np.save(os.path.join(roi_path, 'color.npy'), self.rois[name].color, allow_pickle=True)
            np.save(os.path.join(roi_path, 'filepaths.npy'), self.rois[name].filepaths, allow_pickle=True)
            if self.rois[name].contour_position is not None:
                np.save(os.path.join(roi_path, 'contour_position.npy'),
                        np.array(self.rois[name].contour_position, dtype=object),
                        allow_pickle=True)

    def save_pois(self, path, create_main_folder=False):
        if create_main_folder:
            path = os.path.join(path, 'POIs')
            os.mkdir(path)

        for name in list(self.pois.keys()):
            poi_path = os.path.join(os.path.join(path, name))
            os.mkdir(poi_path)

            np.save(os.path.join(poi_path, 'name.npy'), self.pois[name].name, allow_pickle=True)
            np.save(os.path.join(poi_path, 'visible.npy'), self.pois[name].visible, allow_pickle=True)
            np.save(os.path.join(poi_path, 'color.npy'), self.pois[name].color, allow_pickle=True)
            np.save(os.path.join(poi_path, 'filepaths.npy'), self.pois[name].filepaths, allow_pickle=True)
            np.save(os.path.join(poi_path, 'point_position.npy'), self.pois[name].point_position, allow_pickle=True)

    def load_image(self, image_path, rois=True, pois=True):

        self.array = np.load(os.path.join(image_path, 'array.npy'), allow_pickle=True)
        self.tags = np.load(os.path.join(image_path, 'tags.npy'), allow_pickle=True)
        info = pd.read_pickle(os.path.join(image_path, 'info.p'), )
        for column in list(info.columns):
            setattr(self, column, info.at[0, column])

        if rois:
            roi_names = os.listdir(os.path.join(image_path, 'ROIs'))
            for name in roi_names:
                self.load_rois(os.path.join(image_path, 'ROIs', name))

        if pois:
            roi_names = os.listdir(os.path.join(image_path, 'POIs'))
            for name in roi_names:
                self.load_pois(os.path.join(image_path, 'POIs', name))

    def load_rois(self, roi_path):
        name = str(np.load(os.path.join(roi_path, 'name.npy'), allow_pickle=True))

        existing_rois = list(self.rois.keys())
        if name in existing_rois:
            n = 0
            while n >= 0:
                n += 1
                new_name = name + '_' + str(n)
                if new_name not in existing_rois:
                    name = new_name
                    n = -1

        self.rois[name] = Roi(self)
        self.rois[name].name = name
        self.rois[name].visible = bool(np.load(os.path.join(roi_path, 'visible.npy'), allow_pickle=True))
        self.rois[name].color = list(np.load(os.path.join(roi_path, 'color.npy'), allow_pickle=True))
        self.rois[name].filepaths = str(np.load(os.path.join(roi_path, 'filepaths.npy'), allow_pickle=True))

        if os.path.exists(os.path.join(roi_path, 'contour_position.npy')):
            self.rois[name].contour_position = list(np.load(os.path.join(roi_path, 'contour_position.npy'),
                                                            allow_pickle=True))

    def load_pois(self, poi_path):
        name = str(np.load(os.path.join(poi_path, 'name.npy'), allow_pickle=True))

        existing_pois = list(self.pois.keys())
        if name in existing_pois:
            n = 0
            while n >= 0:
                n += 1
                new_name = name + '_' + str(n)
                if new_name not in existing_pois:
                    name = new_name
                    n = -1

        self.pois[name] = poi(self)
        self.pois[name].name = name
        self.pois[name].visible = bool(np.load(os.path.join(poi_path, 'visible.npy'), allow_pickle=True))
        self.pois[name].color = list(np.load(os.path.join(poi_path, 'color.npy'), allow_pickle=True))
        self.pois[name].filepaths = str(np.load(os.path.join(poi_path, 'filepaths.npy'), allow_pickle=True))

        if os.path.exists(os.path.join(poi_path, 'point_position.npy')):
            self.rois[name].contour_position = list(np.load(os.path.join(poi_path, 'point_position.npy'),
                                                            allow_pickle=True))

    def create_sitk_image(self, empty=False):
        if empty:
            sitk_image = sitk.Image([int(dim) for dim in self.dimensions], sitk.sitkUInt8)
        else:
            sitk_image = sitk.GetImageFromArray(self.array)

        matrix_flat = self.matrix.flatten(order='F')
        sitk_image.SetDirection([float(mat) for mat in matrix_flat])
        sitk_image.SetOrigin(self.origin)
        sitk_image.SetSpacing(self.spacing)

        return sitk_image

    def create_rotated_sitk_image(self, empty=False):
        sitk_image = sitk.GetImageFromArray(self.array)
        matrix_flat = self.matrix.flatten(order='F')
        sitk_image.SetDirection([float(mat) for mat in matrix_flat])
        sitk_image.SetOrigin(self.origin)
        sitk_image.SetSpacing(self.spacing)

        transform = sitk.Euler3DTransform()
        transform.SetRotation(0, 0, 10 * np.pi / 180)
        transform.SetCenter(self.rois['Liver'].mesh.center)
        transform.SetComputeZYX(True)

        resample_image = sitk.ResampleImageFilter()
        resample_image.SetOutputDirection(sitk_image.GetDirection())
        resample_image.SetOutputOrigin(sitk_image.GetOrigin())
        resample_image.SetTransform(transform)
        resample_image.SetInterpolator(sitk.sitkLinear)
        resample_image.Execute(sitk_image)

        # resample_image = sitk.Resample(sitk_image, transform, sitk.sitkLinear, 0.0, sitk_image.GetPixelID())
        return sitk.GetArrayFromImage(resample_image)

    def create_external(self, name='External', color=None, visible=False, filepaths=None, threshold=-250):
        if color is None:
            color = [0, 255, 0]

        if name not in list(self.rois.keys()):
            self.rois[name] = Roi(self, name=name, color=color, visible=visible, filepaths=filepaths)

        mask = external(self.array, threshold=threshold, only_mask=True)
        contours = contours_from_mask(mask.astype(np.uint8))
        positions = self.rois[name].convert_pixel_to_position(pixel=contours)

        self.rois[name].contour_pixel = contours
        self.rois[name].contour_position = positions
        self.rois[name].create_discrete_mesh()

    def compute_aspect(self, slice_plane):
        if slice_plane == 'Axial':
            aspect = np.round(self.spacing[0] / self.spacing[1], 2)
        elif slice_plane == 'Coronal':
            aspect = np.round(self.spacing[0] / self.spacing[2], 2)
        else:
            aspect = np.round(self.spacing[1] / self.spacing[2], 2)

        return aspect

    def compute_bounds(self):
        shape = self.array.shape
        matrix_reshape = self.matrix.reshape(1, 9)[0]
        vtk_image = vtk.vtkImageData()
        vtk_image.SetSpacing(self.spacing)
        vtk_image.SetDirectionMatrix(matrix_reshape)
        vtk_image.SetDimensions([shape[1], shape[2], shape[0]])
        vtk_image.SetOrigin(self.origin)

        x_min, x_max, y_min, y_max, z_min, z_max = vtk_image.GetBounds()

        return [x_min, x_max, y_min, y_max, z_min, z_max]

    def compute_center(self, position=True, zyx=False):
        pixel_index = [int(self.dimensions[2] / 2),
                       int(self.dimensions[1] / 2),
                       int(self.dimensions[0] / 2)]

        if position:
            pixel_to_position_matrix = self.display.compute_matrix_pixel_to_position()
            location = np.asarray([pixel_index[0], pixel_index[1], pixel_index[2], 1])

            center = location.dot(pixel_to_position_matrix.T)[:3]
            if zyx:
                return np.flip(center)
            else:
                return center

        else:
            if zyx:
                return [pixel_index[2], pixel_index[1], pixel_index[0]]
            else:
                return pixel_index

    def compute_corner_positions(self):
        shape = self.array.shape
        matrix_reshape = self.matrix.reshape(1, 9)[0]
        vtk_image = vtk.vtkImageData()
        vtk_image.SetSpacing(self.spacing)
        vtk_image.SetDirectionMatrix(matrix_reshape)
        vtk_image.SetDimensions([shape[1], shape[2], shape[0]])
        vtk_image.SetOrigin(self.origin)

        x_min, x_max, y_min, y_max, z_min, z_max = vtk_image.GetBounds()

        corner_points = [(x_min, y_min, z_min),
                         (x_max, y_min, z_min),
                         (x_max, y_max, z_min),
                         (x_min, y_max, z_min),
                         (x_min, y_min, z_max),
                         (x_max, y_min, z_max),
                         (x_max, y_max, z_max),
                         (x_min, y_max, z_max)]

        return corner_points

    def compute_corner_sides(self):
        corner_points = self.compute_corner_positions()
        points = [corner_points[0], corner_points[4], corner_points[7], corner_points[3],
                  corner_points[1], corner_points[2], corner_points[6], corner_points[5]]
        faces = [4, 0, 1, 2, 3,
                 4, 4, 5, 6, 7,
                 4, 0, 4, 7, 1,
                 4, 3, 2, 6, 5,
                 4, 0, 3, 5, 4,
                 4, 1, 7, 6, 2]

        return pv.PolyData(points, faces)

    def compute_pixel(self, position):
        matrix = copy.deepcopy(self.matrix)

        hold_matrix = np.identity(3, dtype=np.float32)
        hold_matrix[0, :] = matrix[0, :] / self.spacing[0]
        hold_matrix[1, :] = matrix[1, :] / self.spacing[1]
        hold_matrix[2, :] = matrix[2, :] / self.spacing[2]

        position_to_pixel_matrix = np.identity(4, dtype=np.float32)
        position_to_pixel_matrix[:3, :3] = hold_matrix
        position_to_pixel_matrix[:3, 3] = np.asarray(self.origin).dot(-hold_matrix.T)

        location = np.asarray([position[0], position[1], position[2], 1])

        return (np.round(location.dot(position_to_pixel_matrix.T)[:3])).astype(np.int32)

    def compute_position(self, xyz):
        matrix = copy.deepcopy(self.matrix)

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * self.spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * self.spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * self.spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        pixel_to_position_matrix = self.compute_matrix_pixel_to_position()
        location = np.asarray([xyz[0], xyz[1], xyz[2], 1])

        return location.dot(pixel_to_position_matrix.T)[:3]

    def compute_matrix_pixel_to_position(self):
        matrix = copy.deepcopy(self.matrix)
        spacing = self.spacing

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        return pixel_to_position_matrix

    def reset_array(self):
        self.display.secondary_array = None
        self.display.matrix = copy.deepcopy(self.matrix)
        self.display.origin = copy.deepcopy(self.origin)
        self.display.slice_location = self.compute_center(position=False, zyx=True)

    def retrieve_angles(self, order='ZXY'):
        rotation = Rotation.from_matrix(self.display.matrix[:3, :3])

        return rotation.as_euler(order, degrees=True)

    def retrieve_array_plane(self, slice_plane):
        return self.display.compute_array(slice_plane=slice_plane)

    def retrieve_slice_location(self, slice_plane):
        if slice_plane == 'Axial':
            return self.display.slice_location[0]

        elif slice_plane == 'Coronal':
            return self.display.slice_location[1]

        else:
            return self.display.slice_location[2]

    def retrieve_slice_position(self, slice_plane=None):
        pixel_to_position_matrix = self.display.compute_matrix_pixel_to_position()

        if slice_plane is None:
            location = np.asarray([self.display.slice_location[2],
                                   self.display.slice_location[1],
                                   self.display.slice_location[0], 1])
        else:
            if slice_plane == 'Axial':
                location = np.asarray([0, 0, self.display.slice_location[0], 1])
            elif slice_plane == 'Coronal':
                location = np.asarray([0, self.display.slice_location[1], 0, 1])
                print(location)
            else:
                location = np.asarray([self.display.slice_location[2], 0, 0, 1])

        return location.dot(pixel_to_position_matrix.T)[:3]

    def retrieve_scroll_max(self, slice_plane):
        if slice_plane == 'Axial':
            return self.display.scroll_max[0]

        elif slice_plane == 'Coronal':
            return self.display.scroll_max[1]

        else:
            return self.display.scroll_max[2]

    def retrieve_vtk_slice(self, slice_plane):
        return self.display.compute_vtk_slice(slice_plane)

    def retrieve_vtk_volume(self, slice_plane):
        return self.display.compute_vtk_volume(slice_plane)

    def update_rotation(self, r_x=0, r_y=0, r_z=0, base=True):
        if r_x != 0 or r_y != 0 or r_z != 0:
            r = Rotation.from_euler('xyz', [r_x, r_y, r_z], degrees=True)
            new_matrix = r.as_matrix()

            if base:
                base_matrix = copy.deepcopy(self.matrix)
                self.display.matrix = new_matrix @ base_matrix
            else:
                self.display.matrix = new_matrix @ self.display.matrix

            self.display.compute_offaxis_array()
            self.display.compute_scroll_max()
        else:
            self.display.compute_scroll_max()
            self.reset_array()



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:
"""

import numpy as np

from scipy.ndimage import binary_fill_holes
from skimage.measure import label, regionprops


def external(array, threshold=-250, min_volume=100, only_mask=True, less_than=False):
    """
    Identifies the largest connected component in a 3D array (e.g., CT/MRI volume) above or below a threshold,
    optionally returning its mask, centroid positions per slice, number of external components per slice, and bounding
    box. Useful for isolating external anatomy or dominant structures.
    """
    if less_than:
        binary = array < threshold
    else:
        binary = array > threshold
    label_image = label(binary)
    label_regions = regionprops(label_image)
    region_areas = [region.area for region in label_regions]
    max_idx = np.argmax(region_areas)
    bounds = label_regions[max_idx].bbox
    box_image = label_regions[max_idx].image

    mask = np.zeros(array.shape)
    centroid_external = np.zeros((box_image.shape[0], 2))
    external_components = np.zeros((box_image.shape[0], 1))
    for ii in range(box_image.shape[0]):
        filled_image = binary_fill_holes(box_image[ii, :, :])
        fill_image = label(filled_image)
        fill_regions = regionprops(fill_image)
        external_components[ii] = len([region.area for region in fill_regions if region.area > min_volume])

        centroid_external[ii, :] = np.round(np.mean(np.argwhere(filled_image), axis=0))
        mask[ii + bounds[0], bounds[1]:bounds[4], bounds[2]:bounds[5]] = 1 * filled_image

    if only_mask:
        return mask
    else:
        return mask, centroid_external, external_components, bounds
